#!/usr/bin/python3
#
# Swing simulation code for https://zmatt.net/the-physics-of-a-playground-swing/
# Matthew Chapman <matt@zmatt.net>
#
# Model based on https://journals.aps.org/pre/abstract/10.1103/PhysRevE.107.044203
# by Chiaki Hirata, Shun'ichi Kitahara, Yuji Yamamoto, Kazutoshi Gohara, Michael J. Richardson
#

from math import cos,sin,sqrt,pi
import numpy as np
import scipy as sp
import matplotlib.pyplot as plt
import sys

M0=5
L=1.81
l1=0.792
l2=0.393
l3=0.395
m1=31.3
m2=12.3
m3=6.4
M=m1+m2+m3
a=(m2/2+m3)*l2/M
b=(m1+m2/2)*l2/M
N=(M0/2+M)*L
g=9.8
k=0.4
kprime=7/3
I=(M0/3+M)*L*L
Iprime=(m1*l1*l1 + m2*l2*l2 + m3*l3*l3)/3 + (m1*m3-m2*m2/4)*l2*l2/M

theta = -np.radians(15)
phi0 = np.radians(35)
psi0 = np.radians(45)
phimean = np.radians(10)
psimean = np.radians(10)
model = "smooth"
#alpha = np.radians(0) # best for squarewave
#alpha = np.radians(-20) # best for jerky (equivalent to 70 for smooth)
#alpha = np.radians(285) # best for smooth reverse
alpha = np.radians(65) # best for smooth
drag = 1
tstep = 1/120
swingcount = 5
output = "screen"
title = 'Swing simulation (smooth lean)'

fig, ax = plt.subplots(figsize=(6.4,3.6), dpi=300)
ax.set_aspect('equal')
ax.set_xlim([-1.5,1.5])
ax.set_ylim([-2.5,0])
ax.set_xlabel('x (m)')
ax.set_ylabel('y (m)')
ax.set_title(title, fontsize='medium')
fig.tight_layout()

def eval_thetadotdot(theta, thetadot):
  torque = -N*g*sin(theta) \
            + 0.5*m1*l1*g*sin(theta+phi) \
            - 0.5*m3*l3*g*sin(theta+psi) \
            - 0.5*m1*l1*(2/3*l1-L*cos(phi)+a*sin(phi))*phidotdot \
            - 0.5*m3*l3*(2/3*l3+L*cos(psi)+b*sin(psi))*psidotdot \
            - 0.5*m1*l1*(L*sin(phi)+a*cos(phi))*(2*thetadot*phidot+phidot*phidot) \
            - 0.5*m3*l3*(-L*sin(psi)+b*cos(psi))*(2*thetadot*psidot+psidot*psidot) \
            - drag*k*np.sign(thetadot)*(thetadot*thetadot*L*L*L+kprime)
  Ip = m1*l1*(-L*cos(phi)+a*sin(phi))+m3*l3*(L*cos(psi)+b*sin(psi))
  return torque/(I+Iprime+Ip)

def plot_theta(theta):
  ax.plot([1*sin(theta),1.1*sin(theta)], [-1*cos(theta),-1.1*cos(theta)], color='k', linewidth=1)
  ax.text(1.2*sin(theta), -1.2*cos(theta), "$%.0f\\degree$" % np.degrees(theta), fontsize=5, horizontalalignment="center", verticalalignment="center")

ti = 0
plot_theta(theta)
for i in range(swingcount):
  # start of swing, calibrate period
  phi=0
  psi=0
  Ip = m1*l1*(-L*cos(phi)+a*sin(phi))+m3*l3*(L*cos(psi)+b*sin(psi))
  Cs = N-0.5*m1*l1*cos(phi)+0.5*m3*l3*cos(psi)
  Cc = -0.5*m1*l1*sin(phi)+0.5*m3*l3*sin(psi)
  A = abs(theta)
  Tn = 4*sp.special.ellipk(sin(A/2))*sqrt((I+Iprime+Ip)/(g*sqrt(Cs*Cs+Cc*Cc)))
  omega = 2*pi/Tn

  t = 0
  thetadot = 0
  swingdone = False
  while True:
    if model == "squarewave":
      phi = phimean+phi0 if not swingdone else phimean-phi0
      psi = psimean+psi0 if not swingdone else psimean-psi0
      phidot = 0
      psidot = 0
      phidotdot = 0
      psidotdot = 0
    elif model == "jerky":
      phi = phi0*sin(omega*t-alpha+cos(omega*t-alpha)*sin(omega*t-alpha))
      psi = psi0*sin(omega*t-alpha+cos(omega*t-alpha)*sin(omega*t-alpha))
      phidot = 2*phi0*omega*cos(omega*t-alpha+cos(omega*t-alpha)*sin(omega*t-alpha))*(cos(omega*t-alpha)**2)
      psidot = 2*psi0*omega*cos(omega*t-alpha+cos(omega*t-alpha)*sin(omega*t-alpha))*(cos(omega*t-alpha)**2)
      phidotdot = -4*phi0*omega*omega*(sin(omega*t-alpha+cos(omega*t-alpha)*sin(omega*t-alpha))*(cos(omega*t-alpha)**4) + cos(omega*t-alpha+cos(omega*t-alpha)*sin(omega*t-alpha))*sin(omega*t-alpha)*cos(omega*t-alpha))
      psidotdot = -4*psi0*omega*omega*(sin(omega*t-alpha+cos(omega*t-alpha)*sin(omega*t-alpha))*(cos(omega*t-alpha)**4) + cos(omega*t-alpha+cos(omega*t-alpha)*sin(omega*t-alpha))*sin(omega*t-alpha)*cos(omega*t-alpha))
    else: # smooth
      phi = phi0*cos(omega*t-alpha)+phimean
      psi = psi0*cos(omega*t-alpha)+psimean
      phidot = -phi0*omega*sin(omega*t-alpha)
      psidot = -psi0*omega*sin(omega*t-alpha)
      phidotdot = -phi0*omega*omega*cos(omega*t-alpha)
      psidotdot = -psi0*omega*omega*cos(omega*t-alpha)

    if ti % 2 == 0:
      p1 = np.array((L*sin(theta)-a*cos(theta), -L*cos(theta)-a*sin(theta)))
      p2 = np.array((L*sin(theta)+b*cos(theta), -L*cos(theta)+b*sin(theta)))
      p3 = p1 + l1*np.array((-sin(theta+phi), cos(theta+phi)))
      p4 = p2 + l3*np.array((sin(theta+psi), -cos(theta+psi)))
      plot0,=ax.plot([0,L*sin(theta)],[0,-L*cos(theta)],color='black')
      plot1,=ax.plot([p1[0],p2[0]],[p1[1],p2[1]],color='red')
      plot2,=ax.plot([p1[0],p3[0]],[p1[1],p3[1]],color='green')
      plot3,=ax.plot([p2[0],p4[0]],[p2[1],p4[1]],color='blue')
    if ti % 12 == 0:
      mid1 = (p1+p3)/2
      mid2 = (p1+p2)/2
      mid3 = (p2+p4)/2
      cm = (m1*mid1 + m2*mid2 + m3*mid3 + M0*np.array((L/2*sin(theta), -L/2*cos(theta))))/(M+M0)
      plot4,=ax.plot([cm[0]],[cm[1]],marker='o',markersize=1,color='red')
    if ti % 2 == 0:
      if output == "screen":
        plt.draw()
        plt.pause(0.001)
      elif output == "png":
        fig.savefig("frame%04d.png" % (ti/4))
      else: # none
        pass
    if ti % 2 == 0:
      plot0.remove()
      plot1.remove()
      plot2.remove()
      plot3.remove()

    # coupled RK4, see https://math.stackexchange.com/questions/2023819/using-the-runge-kuttas-method-to-solve-a-2nd-derivative-question
    C0 = thetadot
    K0 = eval_thetadotdot(theta, thetadot)
    C1 = thetadot + 0.5*tstep*K0
    K1 = eval_thetadotdot(theta + 0.5*tstep*C0, C1)
    C2 = thetadot + 0.5*tstep*K1
    K2 = eval_thetadotdot(theta + 0.5*tstep*C1, C2)
    C3 = thetadot + tstep*K2
    K3 = eval_thetadotdot(theta + tstep*C2, C3)
    thetadot += tstep/6*(K0+2*K1+2*K2+K3)
    theta += tstep/6*(C0+2*C1+2*C2+C3)

    t = t + tstep
    ti = ti + 1

    if not swingdone and thetadot < 0:
      # right extent of swing
      plot_theta(theta)
      swingdone = True
    if swingdone and theta < 0 and thetadot > 0:
      # left extent of swing
      plot_theta(theta)
      print("period: expected=%f actual=%f" % (Tn, t))
      break

print("theta: %f" % np.degrees(theta))
